/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.healer.genetic;

import static org.junit.Assert.assertFalse;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.UncheckedIOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.Set;

import org.junit.Test;

import Jama.EigenvalueDecomposition;
import de.hft.stuttgart.citydoctor2.check.AbstractCheck;
import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.Checker;
import de.hft.stuttgart.citydoctor2.check.ValidationConfiguration;
import de.hft.stuttgart.citydoctor2.check.error.NonPlanarPolygonDistancePlaneError;
import de.hft.stuttgart.citydoctor2.datastructure.BoundarySurface;
import de.hft.stuttgart.citydoctor2.datastructure.Installation;
import de.hft.stuttgart.citydoctor2.datastructure.CityDoctorModel;
import de.hft.stuttgart.citydoctor2.datastructure.CityObject;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;
import de.hft.stuttgart.citydoctor2.math.CovarianceMatrix;
import de.hft.stuttgart.citydoctor2.math.OrthogonalRegressionPlane;
import de.hft.stuttgart.citydoctor2.math.Plane;
import de.hft.stuttgart.citydoctor2.math.Vector3d;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParseException;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParser;
import de.hft.stuttgart.citydoctor2.parser.InvalidGmlFileException;
import de.hft.stuttgart.citydoctor2.utils.Copy;

public class NonPlanarGenetic {

	private static final int ERROR_PENALTY = 100000;
	private Random r = new Random();
	
	private Map<Integer, List<Long>> durationMap = new HashMap<>(); 

	@Test
	public void findVerticesForNonPlanarTest() throws CityGmlParseException, InvalidGmlFileException, IOException {
		ValidationConfiguration config = ValidationConfiguration.loadStandardValidationConfig();
		config.setNumberOfRoundingPlacesInGlobalParameters(8);
		config.setSchematronFilePathInGlobalParameters(null);
		config.getParserConfiguration().setUseLowMemoryConsumption(false);

		List<Path> gmlFiles = getPathsForGmlFiles();
		for (Path gmlFile : gmlFiles) {
			System.out.println("Parsing File: " + gmlFile);
			CityDoctorModel model = CityGmlParser.parseCityGmlFile(gmlFile.toString(),
					config.getParserConfiguration());
			Checker c = new Checker(config, model);
			c.runChecks();
			List<CheckError> errors = new ArrayList<>();
			model.createFeatureStream().forEach(co -> co.collectContainedErrors(errors));
			assertFalse(errors.isEmpty());
			NonPlanarPolygonDistancePlaneError planarError = null;
			for (CheckError err : errors) {
				if (err instanceof NonPlanarPolygonDistancePlaneError) {
					planarError = (NonPlanarPolygonDistancePlaneError) err;
					break;
				}
			}
			if (planarError == null) {
				throw new IllegalStateException("File " + gmlFile + " does not contain a planarity error");
			}
	
			CityObject feature = getFeature(planarError.getPolygon());
			Polygon polygon = planarError.getPolygon();
			// ensure a gml id has been generated
			polygon.getGmlId();
			int vertexCount = polygon.getExteriorRing().getVertices().size();
			List<Vertex> subList = polygon.getExteriorRing().getVertices().subList(0, vertexCount - 1);

			
			System.out.println("Computing polygon: " + polygon.getGmlId() + " with " + subList.size() + " Vertices");
			FitnessEvaluation<PolygonVertexList> fitnessFunction = createFitnessEvaluationFunction(c, feature);

			// complete calculation of all possibilities
			if (subList.size() == 7) {
				System.out.println();
			}
			Individual<PolygonVertexList> bestOfBoth = null;
			if (subList.size() > 12) {
				System.out.println("Skipping complete calculation, too many vertices");
				bestOfBoth = doGeneticAlgorithmCalculation(polygon, subList, fitnessFunction);
			} else {
				bestOfBoth = doCompleteCalculation(polygon, fitnessFunction, vertexCount, subList);
			}
			
			System.out.println();
			
			// abstands diff, ebene ohne punkt und normale ausgleichsebene
			
//			if (bestOfBoth.getFitness() > -9000) {
				// write only those that do not have errors in neighbor polygons
				// errors in neighbors mean -10000 so check for higher than 9000 to be sure
				try (BufferedWriter writer = Files.newBufferedWriter(Paths.get("result.txt"), StandardOpenOption.APPEND,
						StandardOpenOption.CREATE)) {
					for (int i = 0; i < bestOfBoth.getObject().getUseVertex().length; i++) {
						boolean used = bestOfBoth.getObject().getUseVertex()[i];
						Vertex v = bestOfBoth.getObject().getVertices().get(i);
						int numberOfAdjacentPolygons = 0;
						int numberOfPolygonsWithPlanarityIssues = 0;
						int numberOfPolygonsWithOtherErrors = 0;
						double distanceToPlane = planarError.getPlane().getDistance(v);
						Set<Polygon> adjacentPolygons = v.getAdjacentPolygons(polygon.getParent());
						numberOfAdjacentPolygons = adjacentPolygons.size();
						for (Polygon adjacentPolygon : adjacentPolygons) {
							List<CheckError> adjacentPolygonErrors = new ArrayList<>();
							adjacentPolygon.collectContainedErrors(adjacentPolygonErrors);
							// impossible to have duplicate errors so no Set filter on top
							for (CheckError err : adjacentPolygonErrors) {
								if (err instanceof NonPlanarPolygonDistancePlaneError) {
									numberOfPolygonsWithPlanarityIssues++;
								} else {
									numberOfPolygonsWithOtherErrors++;
								}
							}
							
						}
						if (used) {
							writer.write("1");
						} else {
							writer.write("0");
						}
						
						List<Vertex> vertices = new ArrayList<>();
						for (Vertex vertexForPlan : subList) {
							if (vertexForPlan == v) {
								continue;
							}
							vertices.add(vertexForPlan);
						}
						Vector3d centroid = CovarianceMatrix.getCentroid(vertices);
						EigenvalueDecomposition ed = OrthogonalRegressionPlane.decompose(vertices, centroid);
						Vector3d eigenvalues = OrthogonalRegressionPlane.getEigenvalues(ed);
						Plane plane = OrthogonalRegressionPlane.calculatePlane(centroid, ed, eigenvalues);
						
						double distanceDifferenceWithoutPoint = distanceToPlane - plane.getDistance(v);
						
						writer.write(" 1:");
						writer.write(Integer.toString(numberOfAdjacentPolygons));
						writer.write(" 2:");
						writer.write(Integer.toString(numberOfPolygonsWithPlanarityIssues));
						writer.write(" 3:");
						writer.write(Integer.toString(numberOfPolygonsWithOtherErrors));
						writer.write(" 4:");
						writer.write(Double.toString(distanceToPlane));
						writer.write(" 5:");
						writer.write(Double.toString(distanceDifferenceWithoutPoint));
						writer.write('\n');
					}
				}
//			}
		}
		
		
		for (Entry<Integer, List<Long>> e : durationMap.entrySet()) {
			long sum = 0;
			for (Long l : e.getValue()) {
				sum += l;
			}
			System.out.println(e.getKey() + " Vertices took on average: " + (sum / (float) e.getValue().size()) + "ms");
		}
		
	}

	private Individual<PolygonVertexList> doGeneticAlgorithmCalculation(Polygon polygon, List<Vertex> subList, FitnessEvaluation<PolygonVertexList> fitnessFunction)
			throws IOException {
		Mutation<PolygonVertexList> mutation = createMutationFunction();
		Crossover<PolygonVertexList> crossOver = createCrossoverFunction();
		Heuristic<PolygonVertexList> randomHeuristic = createHeursiticFunction(polygon, subList);
		
		GeneticAlgorithm<PolygonVertexList> algo = GeneticAlgorithm.of(crossOver, mutation, randomHeuristic, fitnessFunction);

		long startGen = System.currentTimeMillis();
		algo.run();
		long deltaGen = System.currentTimeMillis() - startGen;
		System.out.println("Genetic Algo took: " + deltaGen + " ms");
		durationMap.compute(subList.size(), (key, value) -> {
			if (value == null) {
				value = new ArrayList<>();
			}
			value.add(deltaGen);
			return value;
		});
		PolygonVertexList result = algo.getResult().getObject();
		printUsage(result.getUseVertex());
		return algo.getResult();
	}

	private Heuristic<PolygonVertexList> createHeursiticFunction(Polygon polygon, List<Vertex> subList) {
		Heuristic<PolygonVertexList> randomHeuristic = () -> {
			PolygonVertexList vertexList = new PolygonVertexList(polygon, subList);
			boolean[] useVertex = vertexList.getUseVertex();
			for (int i = 0; i < useVertex.length; i++) {
				useVertex[i] = r.nextBoolean();
			}
			validateIndividual(vertexList);
			return vertexList;
		};
		return randomHeuristic;
	}

	private Crossover<PolygonVertexList> createCrossoverFunction() {
		Crossover<PolygonVertexList> crossOver = (p1, p2) -> {
			PolygonVertexList result = new PolygonVertexList(p1.getPolygon(), p1.getVertices());
			int splitIndex = r.nextInt(p1.getUseVertex().length);
			boolean[] array = result.getUseVertex();
			for (int i = 0; i < splitIndex; i++) {
				array[i] = p1.getUseVertex()[i];
			}
			for (int i = splitIndex; i < array.length; i++) {
				array[i] = p2.getUseVertex()[i];
			}
			validateIndividual(result);

			return result;
		};
		return crossOver;
	}

	private Mutation<PolygonVertexList> createMutationFunction() {
		Mutation<PolygonVertexList> mutation = p -> {
			int doSomethingValue = r.nextInt(100);
			if (doSomethingValue > 50) {
				// don't do anything
				return p;
			}
			int index = r.nextInt(p.getUseVertex().length);
			p.getUseVertex()[index] = !p.getUseVertex()[index];

			validateIndividual(p);
			return p;
		};
		return mutation;
	}

	private Individual<PolygonVertexList> doCompleteCalculation(Polygon polygon, FitnessEvaluation<PolygonVertexList> fitnessFunction, int vertexCount,
			List<Vertex> subList) {
		System.out.println("Doing complete calculation");
		long startComplete = System.currentTimeMillis();
		double highestFitness = Double.NEGATIVE_INFINITY;
		int computeCount = 0;
		PolygonVertexList bestIndividual = null;
		for (int i = 7; i < Math.pow(2, vertexCount - 1) - 1; i++) {
			PolygonVertexList individual = new PolygonVertexList(polygon, subList);
			int usedVertexCount = 0;
			boolean[] useVertices = individual.getUseVertex();
			for (int j = 0; j < vertexCount - 1; j++) {
				long l = (i >> j) & 1;
				if (l == 1) {
					usedVertexCount++;
					useVertices[j] = true;
				} else {
					useVertices[j] = false;
				}
			}
			if (usedVertexCount >= 3) {
				double fitnessValue = fitnessFunction.evaluate(individual);
				computeCount++;
				if (fitnessValue > highestFitness) {
					highestFitness = fitnessValue;
					System.out.println("Best Fitness: " + highestFitness);
					bestIndividual = individual;
				}
			} 
			if (usedVertexCount == vertexCount - 1) {
				throw new IllegalStateException("All values should not be computed");
			}
		}
		long deltaComplete =  System.currentTimeMillis() - startComplete;
		System.out.println("End. Best Fitness: " + highestFitness + " Individual: " + bestIndividual);
		System.out.println("Computed " + computeCount + " individuals");
		if (bestIndividual != null) {
			printUsage(bestIndividual.getUseVertex());
		}
		System.out.println("Complete calculation took: " + deltaComplete + "ms");
		durationMap.compute(subList.size(), (key, value) -> {
			if (value == null) {
				value = new ArrayList<>();
			}
			value.add(deltaComplete);
			return value;
		});

		Individual<PolygonVertexList> bestResult = new Individual<PolygonVertexList>(bestIndividual);
		bestResult.setFitness(highestFitness);
		return bestResult;
	}

	private FitnessEvaluation<PolygonVertexList> createFitnessEvaluationFunction(Checker c, CityObject feature) {
		FitnessEvaluation<PolygonVertexList> fitness = p -> {
			List<Vertex> vertices = new ArrayList<>();
			for (int i = 0; i < p.getVertices().size(); i++) {
				if (p.getUseVertex()[i]) {
					vertices.add(p.getVertices().get(i));
				}
			}
			Vector3d centroid = CovarianceMatrix.getCentroid(vertices);
			EigenvalueDecomposition ed = OrthogonalRegressionPlane.decompose(vertices, centroid);
			Vector3d eigenvalues = OrthogonalRegressionPlane.getEigenvalues(ed);
			Plane plane = OrthogonalRegressionPlane.calculatePlane(centroid, ed, eigenvalues);
			// check if it produces errors somewhere else
			// first repair

			// only repair clone
			CityObject copy = Copy.copy(feature);
			Polygon copyP = findPolygonInFeature(copy, p.getPolygon());
			Set<Vertex> allVertices = new HashSet<>();
			// collect all vertices, eliminating duplicates
			collectVertices(copyP, allVertices);
			// project each vertex to the plane and change the coordinates of existing
			// vertices
			for (Vertex v : allVertices) {
				Vector3d newPoint = plane.projectPointToPlane(v);
				v.setX(newPoint.getX());
				v.setY(newPoint.getY());
				v.setZ(newPoint.getZ());
			}
			// update edges
			copy.prepareForChecking();

			// recheck
			c.executeChecksForCheckable(copy);
			List<CheckError> remainingErrors = new ArrayList<>();
			for (Polygon poly : collectNeighborPolygons(copyP)) {
				poly.collectContainedErrors(remainingErrors);
			}
			double distanceSum = 0;
			distanceSum = remainingErrors.size() * ERROR_PENALTY;

			// check the distance of the original vertices to calculate distance
			allVertices = new HashSet<>();
			// collect all vertices, eliminating duplicates
			collectVertices(p.getPolygon(), allVertices);
			for (Vertex v : allVertices) {
				double distance = plane.getDistance(v);
				distanceSum += distance;
			}
			// negative, as higher is better but more distance is worse
			return -distanceSum;
		};
		return fitness;
	}

	public long factorial(int n) {
		long fact = 1;
		for (int i = 2; i <= n; i++) {
			fact = fact * i;
		}
		return fact;
	}

	private List<Path> getPathsForGmlFiles() throws IOException {
		Path entry = Paths.get("C:/Job/CityGML Files/GE_P_NON_PLANAR_POLYGON_DISTANCE_PLANE");
		List<Path> paths = new ArrayList<>();
		Files.list(entry).forEach(path -> collectGmlFiles(path, paths));
		return paths;
	}
	
	private void collectGmlFiles(Path path, List<Path> paths) {
		if (Files.isRegularFile(path) && path.getFileName().toString().endsWith("gml")) {
			paths.add(path);
		} else if (Files.isDirectory(path)) {
			try {
				Files.list(path).forEach(p -> collectGmlFiles(p, paths));
			} catch (IOException e) {
				throw new UncheckedIOException(e);
			}
		}
	}

	private void printUsage(boolean[] useVertex) {
		System.out.println(createUseString(useVertex));
	}

	private String createUseString(boolean[] useVertex) {
		StringBuilder builder = new StringBuilder();
		for (boolean b : useVertex) {
			if (b) {
				builder.append('1');
			} else {
				builder.append('0');
			}
		}
		return builder.toString();
	}

	private void validateIndividual(PolygonVertexList p) {
		int index;
		// count the number of used vertices
		int useCount = 0;
		for (boolean b : p.getUseVertex()) {
			if (b) {
				useCount++;
			}
		}
		if (useCount < 3) {
			// at least 3 points should be used
			for (int i = 0; i < 3 - useCount; i++) {
				index = r.nextInt(p.getUseVertex().length);
				// search for a false index all the time
				while (p.getUseVertex()[index]) {
					index = r.nextInt(p.getUseVertex().length);
				}
				p.getUseVertex()[index] = true;
			}
		} else if (useCount == p.getVertices().size()) {
			// not all vertices should be used
			// flip one randomly back
			index = r.nextInt(p.getUseVertex().length);
			p.getUseVertex()[index] = false;
		}
	}

	private Polygon findPolygonInFeature(CityObject copy, Polygon polygon) {
		Polygon[] result = new Polygon[1];
		copy.accept(new AbstractCheck() {

			@Override
			public void check(Polygon p) {
				if (p.getGmlId().equals(polygon.getGmlId())) {
					result[0] = p;
				}
			}
		});
		return result[0];
	}

	private CityObject getFeature(Polygon polygon) {
		CityObject parent = polygon.getParent().getParent();
		if (parent instanceof BoundarySurface) {
			BoundarySurface bs = (BoundarySurface) parent;
			parent = bs.getParent();
			if (parent instanceof Installation) {
				Installation bi = (Installation) parent;
				parent = bi.getParent();
			}
		} else if (parent instanceof Installation) {
			Installation bi = (Installation) parent;
			parent = bi.getParent();
		}
		return parent;
	}

	private Set<Polygon> collectNeighborPolygons(Polygon polygon) {
		Geometry geometry;
		if (polygon.isLinkedTo()) {
			geometry = polygon.getLinkedFromPolygon().getParent();
		} else {
			geometry = polygon.getParent();
		}
		Set<Polygon> polygons = new HashSet<>();
		for (Vertex v : polygon.getExteriorRing().getVertices()) {
			polygons.addAll(v.getAdjacentPolygons(geometry));
		}

		return polygons;
	}

	private void collectVertices(Polygon p, Set<Vertex> vertices) {
		collectVertices(p.getExteriorRing(), vertices);
		for (LinearRing lr : p.getInnerRings()) {
			collectVertices(lr, vertices);
		}
	}

	private void collectVertices(LinearRing ring, Set<Vertex> vertices) {
		for (Vertex v : ring.getVertices()) {
			vertices.add(v);
		}
	}

}
